package org.rzo.yacron4j;

import io.nettyx.util.HashedWheelTimer;
import io.nettyx.util.Timeout;
import io.nettyx.util.Timer;
import io.nettyx.util.internal.logging.InternalLogger;

import java.text.ParseException;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import org.quartzx.CronExpression;

public class Scheduler
{
	private final Timer _timer;
	private final SchedulerOptions _options;
	private final InternalLogger _logger;
	static AtomicLong ID_COUNTER = new AtomicLong();
	private final boolean _debug;

	public Scheduler()
	{
		this(new SchedulerOptions());
	}

	public Scheduler(SchedulerOptions options)
	{
		_options = (SchedulerOptions) options.clone();
		options.setDebug(true);
		_logger = _options.getLogger();
		_debug = _options.isDebug();

		ExecutorService executor;
		if (options.getMaxThreads() != Integer.MAX_VALUE)
			executor = new ThreadPoolExecutor(
					options.getMaxThreads(),
					options.getMaxThreads(),
					0L,
					TimeUnit.MILLISECONDS,
					new LinkedBlockingQueue<Runnable>(options.getMaxQueueSize()),
					options.getThreadFactory());
		else if (options.getMaxQueueSize() == Integer.MAX_VALUE)
			executor = Executors
					.newCachedThreadPool(options.getThreadFactory());
		else
			executor = new ThreadPoolExecutor(
					options.getMaxQueueSize(),
					options.getMaxThreads(),
					0L,
					TimeUnit.MILLISECONDS,
					new LinkedBlockingQueue<Runnable>(options.getMaxQueueSize()),
					options.getThreadFactory());

		_timer = new HashedWheelTimer(options.getThreadFactory(),
				options.getTickDuration(), TimeUnit.MILLISECONDS,
				options.getTicksPerWheel(), executor);

		((HashedWheelTimer) _timer).start();
	}

	public synchronized ScheduledFuture schedule(Runnable task,
			String cronExpression, TaskOptions options) throws ParseException
	{
		return schedule("noname", task, cronExpression, options);
	}

	public synchronized ScheduledFuture schedule(Callable task,
			String cronExpression, TaskOptions options) throws ParseException
	{
		return schedule("noname", task, cronExpression, options);
	}

	public synchronized ScheduledFuture schedule(Runnable task,
			String cronExpression) throws ParseException
	{
		return schedule("noname", task, cronExpression, new TaskOptions());
	}

	public synchronized ScheduledFuture schedule(Callable task,
			String cronExpression) throws ParseException
	{
		return schedule("noname", task, cronExpression, new TaskOptions());
	}

	public synchronized ScheduledFuture schedule(String id, Runnable task,
			String cronExpression) throws ParseException
	{
		return schedule(id, task, cronExpression, new TaskOptions());
	}

	public synchronized ScheduledFuture schedule(String id, Callable task,
			String cronExpression) throws ParseException
	{
		return schedule(id, task, cronExpression, new TaskOptions());
	}

	public synchronized ScheduledFuture schedule(String id, Runnable task,
			String cronExpression, TaskOptions options) throws ParseException
	{
		if (!_timer.isStarted())
			throw new IllegalStateException("Cannot schedule after shutdown");
		if (task == null)
			throw new RuntimeException("Cannot schedule null task");
		TaskOptions clonedOptions = (TaskOptions) options.clone();
		CronExpression exp = options.isDstConflictIgnore() ? new CronExpression(
				cronExpression) : new DSTCronExpression(cronExpression,
				clonedOptions);
		exp.setTimeZone(options.getTimeZone());
		String uniqueId = id;
		if (id.equals("noname"))
		   uniqueId += "-" + ID_COUNTER.incrementAndGet();
		CronTaskImpl result = new CronTaskImpl(uniqueId, _timer, exp, task,
				clonedOptions, _options.getPersistManager());
		if (_debug)
			_logger.info(uniqueId + ": scheduling " + cronExpression);
		result.start();
		return result;
	}

	public synchronized ScheduledFuture schedule(String id, Callable task,
			String cronExpression, TaskOptions options) throws ParseException
	{
		if (!_timer.isStarted())
			throw new IllegalStateException("Cannot schedule null task");
		if (task == null)
			throw new RuntimeException("task cannot be null");
		TaskOptions clonedOptions = (TaskOptions) options.clone();
		CronExpression exp = options.isDstConflictIgnore() ? new CronExpression(
				cronExpression) : new DSTCronExpression(cronExpression,
				clonedOptions);
		exp.setTimeZone(options.getTimeZone());
		String uniqueId = id + "-" + ID_COUNTER.incrementAndGet();
		CronTaskImpl result = new CronTaskImpl(uniqueId, _timer, exp, task,
				clonedOptions, _options.getPersistManager());
		if (_debug)
			_logger.info(uniqueId + ": scheduling " + cronExpression);
		result.start();
		return result;
	}

	public void shutdown()
	{
		shutdown(true);
	}

	public synchronized void shutdown(boolean mayInterruptIfRunning)
	{
		Set<Timeout> timeouts = _timer.stop();
		if (_debug)
			_logger.info("shutdown cron scheduler: canceling "
					+ timeouts.size() + " tasks");
		for (Timeout timeout : timeouts)
		{
			timeout.cancel();
		}
		_timer.shutdown(mayInterruptIfRunning);
	}

}
